// Copyright (C) 1999-2001 Open Source Telecom Corporation.
//  
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software 
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
// 
// As a special exception to the GNU General Public License, permission is 
// granted for additional uses of the text contained in its release 
// of ccscript.
// 
// The exception is that, if you link the ccscript library with other
// files to produce an executable, this does not by itself cause the
// resulting executable to be covered by the GNU General Public License.
// Your use of that executable is in no way restricted on account of
// linking the ccscript library code into it.
// 
// This exception does not however invalidate any other reasons why
// the executable file might be covered by the GNU General Public License.
// 
// This exception applies only to the code released under the 
// name ccscript.  If you copy code from other releases into a copy of
// ccscript, as the General Public License permits, the exception does
// not apply to the code that you add in this way.  To avoid misleading
// anyone as to the status of such modified files, you must delete
// this exception notice from them.
// 
// If you write modifications of your own for ccscript, it is your choice
// whether to permit this exception to apply to your modifications.
// If you do not wish that, delete this exception notice.  

#ifndef	__CCXX_RTP_H__
#define	__CCXX_RTP_H__

#ifndef __CCXX_SOCKET_H__
#include <cc++/socket.h>
#else
#ifdef	__CCXX_NAMESPACE_H__
#include <cc++/macros.h>
#endif
#endif

typedef enum
{
	RTP_PAYLOAD_PCMU = 0,
	RTP_PAYLOAD_1016,
	RTP_PAYLOAD_G726,
	RTP_PAYLOAD_GSM,
	RTP_PAYLOAD_G723,
	RTP_PAYLOAD_DVI4_8000,
	RTP_PAYLOAD_DVI4_16000,
	RTP_PAYLOAD_LPC,
	RTP_PAYLOAD_PCMA,
	RTP_PAYLOAD_G722,
	RTP_PAYLOAD_L16_DUAL,
	RTP_PAYLOAD_L16_MONO,
	RTP_PAYLOAD_QCELP,

	RTP_PAYLOAD_MPA = 14,
	RTP_PAYLOAD_G728,
	RTP_PAYLOAD_DVI4_11025,
	RTP_PAYLOAD_DVI4_22050,
	RTP_PAYLOAD_G729,

	RTP_PAYLOAD_CELB = 25,
	RTP_PAYLOAD_JPEG,

	RTP_PAYLOAD_NV = 28, 

	RTP_PAYLOAD_H261 = 31,
	RTP_PAYLOAD_MPV,
	RTP_PAYLOAD_MP2T,
	RTP_PAYLOAD_H263,

	RTP_PAYLOAD_INVALID = 128
}	payload_t;

typedef	enum
{
	RTP_PURGE_SEND,
	RTP_PURGE_RECV,
	RTP_PURGE_BOTH
}	rtppurge_t;

class	RTPPacket;

/**
 * A thread serviced packet queue handler for building different kinds of
 * RTP protocol systems.  By making the queue handler a seperate base
 * class it becomes possible to define RTP classes for RTP profiles and
 * sessions of different types.
 *
 * @author David Sugar <dyfet@ostel.com>
 * @short RTP protocol queue handler.
 */
class __EXPORT RTPQueue : protected Thread
{
private:
	RTPPacket *sendfirst, *recvfirst;
	RTPPacket *sendlast, *recvlast;
	Mutex	sendlock, recvlock;
	unsigned short sendseq, recvseq;
	unsigned long sendsources[17];
	unsigned sendcc, segment;
	unsigned char buffer[8192];	// packet receive workspace
	bool marked;
	bool complete;
	TimerPort	timeclock;

	timeout_t	timeout;
	timeout_t	expired;	// time for expired packets

	void Run(void);
	
protected:
	struct timeval starttimer;
	volatile bool active;

	/**
	 * A plugin for RTCP packet scheduling based on the number of
	 * contributing sources in a RTP session.  The default interval
	 * is to use 5 seconds for posting RTCP reports.  If no RTCP
	 * interface is used, this should return 0.
	 *
	 * @return number of seconds in RTCP intervals.
	 */
	virtual time_t getControlTimer(void)
		{return 0;};

	/**
	 * A plugin point for picking up incoming RTCP packets if they
	 * are waiting.  This is used as a simple "check"; if packets
	 * are pending, they are processed in the derived class.  A
	 * timeout for the maximum interval since the last RTCP packet
	 * had been received is also returned.  This is checked once a
	 * second.
	 *
	 * @return number of seconds since last RTCP arrival.
	 */
	virtual unsigned getControl(void)
		{return 0;};


	/**
	 * A plugin point for posting of BYE messages.
	 */
	virtual void Bye(void)
		{return;};
	
	/**
	 * A plugin point for posting of RTCP messages.
	 */
	virtual void putControl(void)
		{return;};

	/**
	 * A plugin point for timer tick driven events.
	 */
	virtual void timerTick(void)
		{return;};

	/**
	 * This computes the timeout period for scheduling transmission
	 * of the next packet at the "head" of the send buffer.  If no
	 * packets are waiting, a default timeout is used.  This actually
	 * forms the "isPending()" timeout of the rcp receiver in the
	 * service thread.
	 *
	 * @return timeout until next packet is scheduled to send.
	 */
	timeout_t getTimeout(void);

	/**
	 * This function is used to check for and schedule against
	 * arriving packets based on the derived connection type.
	 *
	 * @return true if packet waiting for processing.
	 * @param number of milliseconds to wait.
	 */
	virtual bool isPendingArrival(timeout_t timeout) = 0;

	/**
	 * This function is used by the service thread to process
	 * the next outgoing packet pending in the send queue.
	 *
	 * @return number of bytes sent.  0 if silent, <0 if error.
	 */
	int sendPacket(void);

	/**
	 * This function is used to purge the packet queue(s).
	 */
	void Purge(rtppurge_t flag);

	/**
	 * This function performs the physical I/O for writing a
	 * packet to the destination.  It is a virtual that is
	 * overriden in the derived class.
	 *
	 * @return number of bytes sent.
	 * @param packet to write.
	 * @param length of data to write.
	 */
	virtual int writePacket(unsigned char *packet, unsigned len) = 0;

	/**
	 * This function is used by the service thread to process
	 * the next incoming packet and place it in the receive list.
	 *
	 * @return number of payload bytes received.  <0 if error.
	 */
	int recvPacket(void);

	/**
	 * A virtual function to support parsing of arriving packets
	 * to determine if they should be kept in the queue and to
	 * dispatch events.
	 *
	 * @return true if packet is kept.
	 * @param packet returned.
	 */
	virtual bool gotPacket(RTPPacket *packet)
		{return true;};

	/**
	 * This function performs the physical I/O for reading a
	 * packet from the source.  It is a virtual that is
	 * overriden in the derived class.
	 *
	 * @return number of bytes read.
	 * @param packet read buffer.
	 * @param length of data to read.
	 */
	virtual int readPacket(unsigned char *buffer, unsigned len) = 0;

	/**
	 * This is used to fetch a packet in the receive queue and to
	 * expire packets older than the current timestamp.
	 *
	 * @return packet buffer object for current timestamp if found.
	 * @param timestamp requested.
	 */
	RTPPacket *getWaiting(unsigned long timestamp);

	/**
	 * This function ends the queues and service threads.
	 */
	void endQueue(void);
	
public:
	RTPQueue(int pri);
	virtual ~RTPQueue()
		{endQueue();};

	/**
	 * This is used to create a data packet in the send queue.  
	 * Sometimes a "NULL" or empty packet will be used instead, and
	 * these are known as "silent" packets.  "Silent" packets are
	 * used simply to "push" the scheduler along more accurately
	 * by giving the appearence that a next packet is waiting to
	 * be sent and to provide a valid timestamp for that packet.
	 *
	 * @param payload format of this packet.
	 * @param timestamp for expected send time of packet.
	 * @param data value or NULL if special "silent" packet.
	 * @param length, may be 0 to indicate a default by payload type.
	 */
	void putPacket(unsigned long stamp, payload_t payload, unsigned char *data = NULL, unsigned len = 0);

	/**
	 * Retreive data from a specific timestamped packet if such a
	 * packet is currently available in the receive buffer.
	 *
	 * @return number of packet data bytes retrieved.
	 * @param timestamp of packet desired.
	 * @param data buffer to copy into.
	 * @param maximum data size.
	 */
	unsigned getPacket(unsigned long stamp, unsigned char *data, unsigned max);

	/**
	 * Get the payload type of a specific packet by timestamp.
	 *
	 * @return payload of specified packet if found.
	 * @param timestamp to find.
	 */
	payload_t getPayload(unsigned long timestamp);

	/**
	 * Set partial data for an already queued packet.  This is often
	 * used for multichannel data.
	 *
	 * @return number of packet data bytes set.
	 * @param timestamp of packet desired.
	 * @param data buffer to copy from.
	 * @param offset to copy from.
	 * @param maximum data size.
	 */
	unsigned setPartial(unsigned long timestamp, unsigned char *data, unsigned offset, unsigned max);

	/**
	 * Get partial data from a packet.  This is often used to support
	 * oddball hardware that has unusual or non-standard codec
	 * buffering intervals.
	 *
	 * @return number of packet data bytes retrieved.
	 * @param timestamp of packet desired.
	 * @param data buffer to copy into.
	 * @param offset to copy from.
	 * @param maximum data size.
	 */
	unsigned getPartial(unsigned long timestamp, unsigned char *data, unsigned int offset, unsigned int max);

	/**
	 * Get active connection state flag.
	 *
	 * @return true if connection "active".
	 */
	inline bool isActive(void)
		{return active;};

	/**
	 * Set the default scheduling timeout to use when no data
	 * packets are waiting to be sent.
	 *
	 * @param timeout in milliseconds.
	 */
	inline void setTimeout(timeout_t t)
		{timeout = t;};

	/**
	 * Set the "expired" timer for expiring packets pending in
	 * the send queue which have gone unsent and are already
	 * "too late" to be sent now.
	 *
	 * @param timeout to expire unsent packets in milliseconds.
	 */
	inline void setExpired(timeout_t t)
		{expired = t;};

	/**
	 * Set maximum packet segment size before fragmenting sends.
	 *
	 * @param maximum packet size.
	 */
	inline void setSegmentSize(unsigned size)
		{segment = size;};

	/**
	 * Is last packet processed "complete" or are sequences
	 * missing?
	 *
	 * @return true if complete.
	 */
	inline bool isComplete(void)
		{return complete;};

	/**
	 * Is last packet processed "marked" (used to signal end on
	 * multi-sequence sends).
	 *
	 * @return true if marked.
	 */
	inline bool isMarked(void)
		{return marked;};

	/**
	 * Determine if packets are waiting in the queue.
	 *
	 * @return true if packets are waiting.
	 */
	bool isWaiting(void);

	/**
	 * Determine if outgoing packets are waiting to send.
	 *
	 * @return true of packets waiting to send.
	 */
	bool isSending(void);

	/**
	 * Get timestamp of first packet waiting in the queue.
	 *
	 * @return timestamp of first arrival packet.
	 */
	unsigned long getFirstTimestamp(void);

	/**
	 * Get the sequence id of the first packet waiting.
	 *
	 * @return sequence id of first packet.
	 */
	unsigned short getFirstSequence(void);

	/**
	 * This is used to initialize the packet send sequence id.
	 *
	 * @param sequence id to start from.
	 */
	void setSequence(unsigned short seq);

	/**
	 * Set the packet timeclock for synchronizing timestamps.
	 */
	inline void setTimeclock(void)
		{timeclock.setTimer();};

	/**
	 * Get the packet timeclock for synchronizing timestamps.
	 *
	 * @return runtime in milliseconds since last set.
	 */
	inline timeout_t getTimeclock(void)
		{timeclock.getElapsed();};
};

/**
 * RTCP interface class.  This drives RTCP related session control.
 *
 * @author David Sugar <dyfet@ostel.com>
 * @short RTCP packet control interface.
 */
class __EXPORT RTCPSocket : private UDPSocket
{
protected:
	RTCPSocket(const InetAddress &bind, tpport_t port);

	sockerror_t Connect(const InetHostAddress &ia, tpport_t port);

	inline bool isPendingControl(timeout_t timeout)
		{return isPending(SOCKET_PENDING_INPUT, timeout);};

	inline int getControl(unsigned char *buffer, size_t len)
		{return ::read(so, buffer, len);};

	inline int putControl(unsigned char *buffer, size_t len)
		{return ::send(so, buffer, len, MSG_DONTWAIT);};

	/**
	 * A plugin point for hello message.
	 */
	virtual void gotHello(const char *sdes)
		{return;};

	/**
	 * A plugin point for goodbye message.
	 */
	virtual void gotGoodbye(void)
		{return;};
		
public:
	inline void endControl(void)
		{endSocket();};
};

/**
 * Generic RTP protocol stack for exchange of realtime data.  This
 * stack uses the concept of packet send and receive queues to schedule
 * and buffer outgoing packets and to arrange or reorder incoming packets
 * as they arrive.  A single service thread both schedules sending of
 * outgoing packets and receipt of incoming packets.
 *
 * @author David Sugar <dyfet@ostel.com>
 * @short RTP protocol stack based on Common C++.
 */
class __EXPORT RTPSocket : public RTPQueue, protected UDPReceive, public UDPTransmit, protected RTCPSocket
{
protected:

	tpport_t base;

	bool isPendingArrival(timeout_t timeout)
		{return UDPReceive::isPendingReceive(timeout);};

	int readPacket(unsigned char *buffer, unsigned len)
		{return Receive(buffer, len);};

	int writePacket(unsigned char *buffer, unsigned len)
		{return Transmit((const char *)buffer, len);};

public:
	RTPSocket(const InetAddress &bind, tpport_t port = 5004, int pri = 0);
	~RTPSocket();
	void endSocket(void);

	/**
	 * Connect to a foriegn RTP socket and start the service
	 * thread.  If no port is specified then it is assumed to
	 * be the same as the locally bound port number.
	 *
	 * @return failure type.
	 * @param address of foreign socket.
	 * @param port number of foreign connection.
	 */
	sockerror_t Connect(const InetHostAddress &ia, tpport_t port = 0);
};


/**
 * A peer associated RTP socket pair for physically connected peer
 * hosts.  This has no RTCP and assumes the receiver is connected
 * to a known transmitter, hence no "foreign" packets will arrive.
 *
 * @author David Sugar
 * @short RTP peer host over UDP.
 */
class __EXPORT RTPDuplex : public RTPQueue, protected UDPReceive, public UDPTransmit
{
private:
	tpport_t base;
protected:
	bool isPendingArrival(timeout_t timeout)
		{return isPendingReceive(timeout);};

	int writePacket(unsigned char *buffer, unsigned len)
		{return Transmit((const char *)buffer, len);};

	int readPacket(unsigned char *buffer, unsigned len)
		{return Receive(buffer, len);};

public:
	RTPDuplex(const InetAddress &bind, tpport_t local, tpport_t remote, int pri);
	~RTPDuplex();

	sockerror_t Connect(const InetHostAddress &host, tpport_t port = 0);
};

#endif

